import type {
  ComponentRequest,
  RefreshSource,
  TransactionOptions,
  TransactionResponse,
  UUID,
} from '../types';
import { seal, unseal } from './encryption';

export type TransactionType = 'withdraw' | 'deposit';
export type ConnectionStatus = 'active' | 'completed' | 'closed';
export interface ConnectionData<T extends ComponentRequest> {
  tx: T;
  authToken?: string;
  txOptions?: TransactionOptions;
  result?: TransactionResponse<T>;
}

export interface ConnectionMetadata {
  nodeUrl?: string;
  participants: string[];
  transactionType: TransactionType;
}
export interface ConnectionResponse<T extends ComponentRequest> {
  id: UUID;
  version: number;
  status: ConnectionStatus;
  metadata: ConnectionMetadata;
  data: ConnectionData<T>;
  key: string;
}

export interface ConnectionResult<T extends ComponentRequest> {
  id: UUID;
  metadata: ConnectionMetadata;
  status: ConnectionStatus;
  tx?: T;
  result?: TransactionResponse<T>;
}

export async function getRefreshResult<T extends ComponentRequest>(
  refreshSource: RefreshSource,
): Promise<ConnectionResult<T>> {
  const response = await fetch(refreshSource.url, {
    method: 'GET',
  });

  if (!response.ok) {
    throw new Error(`Failed to get connection: ${await response.text()}`);
  }

  const result = await response.json();

  if (!result.id || !result.metadata || !result.status || !result.sealed) {
    throw new Error('Data missing from server response');
  }

  const basicData = {
    id: result.id,
    metadata: result.metadata,
    status: result.status,
  };

  if (result.status === 'closed') {
    return basicData;
  }

  // Get the latest sealed data
  const latestSealed = result.sealed[result.sealed.length - 1];

  // Decrypt the data
  const data = await unseal<ConnectionData<T>>({
    ciphertext: latestSealed,
    key: refreshSource.key,
  });

  if (result.status === 'completed') {
    return {
      ...basicData,
      result: data.result,
    };
  }

  return {
    ...basicData,
    tx: data.tx,
  };
}

/**
 * Manages encrypted connections using Cloudflare Durable Objects
 */
export class ConnectionManager {
  private endpoint: string;

  constructor(endpoint: string) {
    this.endpoint = endpoint;
  }

  /**
   * Creates a new encrypted connection
   * @param data The component request data to encrypt and store
   * @param participants Array of participant identifiers
   * @returns Promise resolving to connection details including ID, version, and encryption key
   */
  async create<T extends ComponentRequest>(
    data: ConnectionData<T>,
    metadata: ConnectionMetadata,
  ): Promise<ConnectionResponse<T>> {
    // Encrypt the data
    const sealed = await seal(data);

    // Prepare the request body
    const body = {
      metadata,
      sealed: sealed.ciphertext,
    };

    // Create the connection
    const response = await fetch(this.endpoint, {
      method: 'POST',
      headers: {
        'Content-Type': 'application/json',
      },
      body: JSON.stringify(body),
    });

    if (!response.ok) {
      throw new Error(`Failed to create connection: ${await response.text()}`);
    }

    const result = await response.json();

    return {
      id: result.id,
      version: result.version,
      status: result.status,
      metadata: metadata,
      data: data,
      key: sealed.key,
    };
  }

  /**
   * Updates an existing connection with new encrypted data
   * @param id Connection ID
   * @param data New data to encrypt and store
   * @param version Current version number
   * @returns Promise resolving to updated connection details including new encryption key
   */
  async update<T extends ComponentRequest>(
    id: UUID,
    data: ConnectionData<T>,
    version: number,
    status: ConnectionStatus,
    key: string,
  ): Promise<ConnectionResponse<T>> {
    // Encrypt the new data
    const sealed = await seal(data, key);

    // Prepare the request body
    const body = {
      sealed: sealed.ciphertext,
      version,
      status,
    };

    // Update the connection
    const response = await fetch(`${this.endpoint}/${id}`, {
      method: 'PATCH',
      headers: {
        'Content-Type': 'application/json',
      },
      body: JSON.stringify(body),
    });

    if (!response.ok) {
      throw new Error(`Failed to update connection: ${await response.text()}`);
    }

    const result = await response.json();

    return {
      id: result.id,
      metadata: result.metadata,
      version: result.version,
      status: result.status,
      data: data,
      key: sealed.key,
    };
  }

  /**
   * Retrieves and decrypts connection data
   * @param id Connection ID
   * @param key Encryption key from previous create/update operation
   * @returns Promise resolving to connection details including decrypted data
   */
  async get<T extends ComponentRequest>(
    id: UUID,
    key: string,
  ): Promise<ConnectionResponse<T>> {
    // Get the connection data
    const response = await fetch(`${this.endpoint}/${id}`, {
      method: 'GET',
    });

    if (!response.ok) {
      throw new Error(`Failed to get connection: ${await response.text()}`);
    }

    const result = await response.json();

    // Get the latest sealed data
    const latestSealed = result.sealed[result.sealed.length - 1];

    // Decrypt the data
    const data = await unseal<ConnectionData<T>>({
      ciphertext: latestSealed,
      key,
    });

    return {
      id: result.id,
      status: result.status,
      version: result.version,
      metadata: result.metadata,
      data,
      key,
    };
  }

  /**
   * Closes a connection
   * @param id Connection ID
   */
  async close(id: UUID): Promise<void> {
    const response = await fetch(`${this.endpoint}/${id}`, {
      method: 'DELETE',
    });

    if (!response.ok) {
      throw new Error(`Failed to close connection: ${await response.text()}`);
    }
  }
}
